/** smtpd.c
*
*/
char version_txt[] =

    "$VER: SMTPd 1.5 (19.8.99)";
/*         ===================
*
*        SMTP mail server.
*/
static char *usage =
   "Usage: smtpd [-h] [-m <mail-dir>] [-u] [-d <info-level>] [-s <sound-file>]\n"
   "where:\n"
   "-h   gives this message\n"
   "-m   sets the incoming mail directory (default: /boot/home/mail/in)\n"
   "-u   turns off checking if the user exists on the system\n"
   "-d   turns on debugging information, with info levels 0 through 7\n"
   "-s   sets a soundfile to use when mail arrives\n";
/*
*
*   By W.G.J Langeveld.
*
**/
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <sys/dir.h>
#include <pwd.h>

#include <sys/socket.h>

#include <TypeConstants.h>

#include <fs_attr.h>
#include <parsedate.h>

#include "mailbeep.h"
#include "setmimetype.h"
#include "readn.h"
/*
*   Prototypes.
*/
void dprintf(long, char *, ...);
long putline(long, char *, ...);
long getline(long, char *);
long extractpath(char *, char *);
long collectfile(long, char *, char *, char *, char *, char *);
void SetParsedAttrs(FILE *fp);
char *findopt(int argc, char **argv, char opt);
/*
*   Info levels:
*/
#define SMERR_FATAL    0
#define SMERR_SEVERE   1
#define SMERR_ERROR    2
#define SMERR_WARNING  3
#define SMERR_INFO     4
#define SMERR_DETAIL   6

static long infolevel = SMERR_SEVERE;
/*
*   Some defines
*/
#define MAXHOSTNAMELEN   64
#define MAXPATHLEN     4096
#define MAXLINE        4096
/*
*   Commands known by this server
*/
#define MKL(a,b,c,d) (((a) << 24) | ((b) << 16) | ((c) << 8) | (d))

#define HELO MKL('H','E','L','O')
#define MAIL MKL('M','A','I','L')
#define RCPT MKL('R','C','P','T')
#define DATA MKL('D','A','T','A')
#define RSET MKL('R','S','E','T')
#define SEND MKL('S','E','N','D')
#define SOML MKL('S','O','M','L')
#define SAML MKL('S','A','M','L')
#define VRFY MKL('V','R','F','Y')
#define EXPN MKL('E','X','P','N')
#define HELP MKL('H','E','L','P')
#define NOOP MKL('N','O','O','P')
#define QUIT MKL('Q','U','I','T')
#define TURN MKL('T','U','R','N')
/*
*   DBUG is not part of the spec. It will generate a "not implemented" error,
*   but will in fact set the "infolevel" to a high value.
*/
#define DBUG MKL('D','B','U','G')

/**
*
*   Main program. Executed from inetd.
*
**/
int main(int argc, char **argv)
{
   long  ss = -1, ll, s;
   int port, n;
   struct sockaddr_in sin, sin2;
   char *maildir     = NULL;
   char *help        = NULL;
   char *nousercheck = NULL;
   char *debug       = NULL;
   char *soundfile   = NULL;

   long len, mailnum, started, idx;

   char myhostname[MAXHOSTNAMELEN + 4];
   char user[MAXHOSTNAMELEN + 4];
   char hostname[MAXHOSTNAMELEN + 4];
   char fpath[MAXPATHLEN + 4];
   char rpath[MAXPATHLEN + 4];
   char bf[MAXLINE + 16], *buff, *ptr1;

   struct hostent *he;
   struct passwd  *pw;
/*
*   Arguments
*/
   help         = findopt(argc, argv, 'h');
   if (help) {
      printf(usage);
      exit(0);
   }

   maildir      = findopt(argc, argv, 'm');
   if (maildir == NULL) maildir = "/boot/home/mail/in";

   nousercheck  = findopt(argc, argv, 'u');
   debug        = findopt(argc, argv, 'd');
   soundfile    = findopt(argc, argv, 's');
/*
*   Get our own host name and user name.
*/
   gethostname(myhostname, MAXHOSTNAMELEN);
   if (myhostname[0] == '\0') {
      dprintf(SMERR_FATAL, "Smtpd: can't find local host name\n");
      goto cleanup;
   }

   he = gethostbyname(myhostname);
   strcpy(myhostname, he->h_name);

   if (myhostname[0] == '\0') {
      dprintf(SMERR_FATAL, "Smtpd: invalid local host name\n");
      goto cleanup;
   }
/*
*   Done with the preliminaries, now initialize our various buffers
*/
   buff = &bf[5];

   user[0]     = 0;
   hostname[0] = 0;
   fpath[0]    = 0;
   rpath[0]    = 0;
   bf[0]       = 0;
   buff[0]     = 0;
/*
*   Check for info level flag
*/
   if (debug) infolevel = atol(debug);

   if (infolevel < 0) infolevel = 0;
   if (infolevel > 7) infolevel = 7;
/*
*   Set up TCP server
*/
   ss = socket(AF_INET, SOCK_STREAM, 0);
   if (ss < 0) {
      dprintf(SMERR_FATAL, "Unable to obtain a socket\n");
      exit(20);
   }
/*
*   We're an SMTP server.
*/ 
   port = 25;

   sin.sin_family      = AF_INET;
   sin.sin_addr.s_addr = INADDR_ANY;
   sin.sin_port        = htons(port);

   if (bind(ss, (struct sockaddr *) &sin, sizeof (sin)) < 0) {
      dprintf(SMERR_FATAL, "%ld> Unable to bind\n", s);
      exit(20);
   }

   if (listen(ss, 5) < 0) {
      dprintf(SMERR_FATAL, "%ld> Listen failure\n", s);
      exit(20);
   }

   for (;;) {
      n = sizeof(sin2);

      s = accept(ss, (struct sockaddr *) &sin2, &n);
      if (s < 0) {
         dprintf(SMERR_FATAL, "%ld> Accept failure\n", ss);
         exit(20);
      }
      dprintf(SMERR_INFO, "%s %ld: START\n", &version_txt[6], s);
/*
*   Start SMTP protocol
*/
      putline(s, "220 %s - BeOS System %s Ready", myhostname, &version_txt[6]);
      while (1) {
         n = getline(s, bf);
         if (n < 0) break;
/*
*   Find which command is in the first four characters
*/
         if (n < 4) {
            idx = -1;
         }
         else {
            bf[4] = '\0';
            for (ll = 0; ll < 4; ll++) bf[ll] = toupper(bf[ll]);
            idx = *(unsigned long *) bf;
            idx = ntohl(idx); 
            len = strlen(buff);
         }
/*
*   Handle each command
*/
         switch (idx) {
            case HELO :
               user[0]     = 0;
               hostname[0] = 0;
               fpath[0]    = 0;
               rpath[0]    = 0;

               if (len == 0) {
                  putline(s, "501 No host name");
               }
               else if (len > 64) {
                  putline(s, "500 Line too long");
               }
               else {
                  strcpy(hostname, buff);
                  if (ptr1 = strchr(hostname, '\r')) *ptr1 = 0;
                  if (ptr1 = strchr(hostname, '\n')) *ptr1 = 0;
                  started = 1;
                  putline(s, "250 %s", myhostname);
               }
               break;
            case MAIL :
               if (extractpath(rpath, buff)) started = 2;
               putline(s, buff);
               break;
            case RCPT :
               if (started < 2) {
                  putline(s, "503 No valid MAIL command received");
               }
               else if (started > 2) {
                  putline(s, "452 Single-user system, multiple recipients illogical");
               }
               else {
                  if (extractpath(fpath, buff)) {
                     strcpy(user, fpath);
                     ptr1 = strchr(user, '@');
                     if (ptr1 == NULL) {
                        strcpy(buff, "553 Incorrect mail box syntax");
                     }
                     else {
                        *ptr1++ = 0;
                        if (strcasecmp(ptr1, myhostname) == 0) {
                           if (nousercheck ||
                               ((pw = getpwnam(user)) && (strcasecmp(pw->pw_name, user) == 0))) {
                              strcpy(buff, "250 OK");
                              started = 3;
                           }
                           else {
                              strcpy(buff, "550 No such user");
                           }
                        }
                        else {
                           strcpy(buff, "551 User not local");
                        }
                     }
                  }
                  putline(s, buff);
               }
               break;
            case DATA :
               if (started != 3) {
                  putline(s, "503 No recipient");
               }
               else {
                  putline(s, "354 OK");

                  if (collectfile(s, maildir, myhostname, rpath, hostname, buff)) {
                     putline(s, "250 OK");
                     mailbeep(soundfile);
                  }
                  else {
                     putline(s, "554 Error collecting data");
                  }
                  started = 2;
               }
               break;
            case RSET :
               user[0]     = 0;
               hostname[0] = 0;
               fpath[0]    = 0;
               rpath[0]    = 0;
               bf[0]       = 0;
               buff[0]     = 0;
               started     = 0;
/* Fall through */
            case NOOP :
               putline(s, "250 OK");
               break;
            case DBUG :
               infolevel = atol(buff);

               if (infolevel < 0) infolevel = 0;
               if (infolevel > 7) infolevel = 7;
/* Fall through */
            case SEND :
            case SOML :
            case SAML :
            case VRFY :
            case EXPN :
            case TURN :
               putline(s, "502 Command not implemented");
               break;
            case HELP :
               putline(s, "211-Available Commands:");
               putline(s, "211 DATA  HELO  HELP  MAIL  NOOP  QUIT  RCPT  RSET");
               break;
            case QUIT :
               putline(s, "221 %s Service closing transmission channel", myhostname);
               goto cleanup;
               break;
            default   :
               putline(s, "502 Command not recognized");
               break;
         }
      }
cleanup:
   }

   return;
}


/**
*
*   Find an option in the optiob array.
*
**/
char *findopt(int argc, char **argv, char opt)
{
   int i;
   for (i = 1; i < argc; i++) {
      if (argv[i][0] == '-') {
         if (argv[i][1] == opt) {
            if (argv[i][2] != '\0') {
               return(&argv[i][2]);
            }
            else {
               i++;
               if (argc > i) return(argv[i]);
               else          return("");
            }
         }
      }
   }
   return(NULL);
}


/**
*
*   Get a line from remote
*
**/
long getline(long s, char *buf)
{
   long n;

   n = readline(s, buf, MAXLINE);
   if (n >= 0) {
      dprintf(SMERR_INFO,  "Smtpd %ld- R: %s", s, buf);
   }
   else {
      dprintf(SMERR_WARNING, "Smtpd %ld: remote terminated early\n", s);
   }

   return(n);
}

/**
*
*   Send a line to the remote
*
**/
long putline(long s, char *fmt, ...)
{
   va_list args;
   int len;
   char buf[260];

   if (fmt) {
      va_start(args, fmt);
      vsprintf(buf, fmt, args);

      dprintf(SMERR_INFO, "Smtpd %ld- S: %s\n", s, buf);

      len = strlen(buf);
      buf[len++] = '\r';
      buf[len++] = '\n';

      writen(s, buf, len);

      va_end(args);
   }
}

/**
*
*   Extract a path specification
*
**/
long extractpath(char *s, char *t)
{
   char *ptr1, *ptr2;
   long len, retval = 0;

   *s = '\0';

   ptr1 = strchr(t, '<');
   ptr2 = strchr(t, '>');

   if ((ptr1 == NULL) || (ptr2 == NULL)) {
      strcpy(t, "501 Illegal path");
   }
   else {
      ptr1++;
      *ptr2 = '\0';
      len = strlen(ptr1);
      if (len > 256) {
         strcpy(t, "500 Line too long");
      }
      else {
         strcpy(s, ptr1);
         strcpy(t, "250 OK");
         retval = 1;
      }
   }

   return(retval);
}

/**
*
*   Collect the mail message
*
**/
long collectfile(long s, char *maildir, char *myhostname,
                         char *rpath, char *hostname, char *buff)
{
   register int len, retval = 1;
   char *line;
   FILE *fp;
   char datebuf[80];
   struct Window *oldw;
   unsigned long number;
   unsigned long seqno = 0;
   char filename[1024];
/*
*   Create the file with a unique name.
*/
   len = strlen(maildir);
   number = time(NULL);

   while (1) {
      sprintf(filename, "%s%s%s_%ld_%ld", maildir,
              ((maildir[len - 1] != '/') && (maildir[len - 1] != ':')) ? "/" : "",
              rpath, number, seqno);

      dprintf(SMERR_DETAIL, "Smtpd %ld: opening file %s\n", s, filename);
      if (access(filename, F_OK) != 0) break;

      seqno++;
   }

   fp = fopen(filename, "w+");
   if (fp == NULL) return(0);

   sprintf(buff, "Return-Path: <%s>\r\n", rpath);
   fwrite(buff, 1, strlen(buff), fp);
   dprintf(SMERR_DETAIL, buff);

   strftime(datebuf, sizeof(datebuf), "%a, %d %b %Y %H:%M:%S %Z", localtime(&number));

   sprintf(buff, "Received: from %s by %s with\r\n   BeOS SMTPd#%05ld; %s\n",
                      hostname, myhostname, number, datebuf);

   fwrite(buff, 1, strlen(buff), fp);
   dprintf(SMERR_DETAIL, buff);

   while (1) {
/*
*   Our "readline" turns cr/lf into lf.
*/
      readline(s, buff, MAXLINE);
      if (buff[0] == '.') {
         if (buff[1] == '\n') break;
         line = &buff[1];
      }
      else {
         line = buff;
      }

      len = strlen(line);
/*
*   Some mail readers on BeOS need the cr/lf pair, so put it back in
*/
      line[len - 1] = '\r';
      line[len++]   = '\n';
      if (fwrite(line, 1, len, fp) != len) retval = 0;
   }

   if (retval == 0) dprintf(SMERR_WARNING, "Smtpd %ld: write error", s);

   set_mime_type(filename, "text/x-email");

   SetParsedAttrs(fp);

   fclose(fp);

   return(retval);
}

/**
*
*   Copied from E-mail.h (can't include it, too much C++ stuff).
*
**/
#define B_MAIL_ATTR_NAME	"MAIL:name"		/* indexed string*/
#define B_MAIL_ATTR_STATUS	"MAIL:status"		/* indexed string*/
#define B_MAIL_ATTR_PRIORITY	"MAIL:priority"		/* indexed string*/
#define B_MAIL_ATTR_TO		"MAIL:to"		/* indexed string*/
#define B_MAIL_ATTR_CC		"MAIL:cc"		/* indexed string*/
#define B_MAIL_ATTR_FROM	"MAIL:from"		/* indexed string*/
#define B_MAIL_ATTR_SUBJECT	"MAIL:subject"		/* indexed string*/
#define B_MAIL_ATTR_REPLY	"MAIL:reply"		/* indexed string*/
#define B_MAIL_ATTR_WHEN	"MAIL:when"		/* indexed time*/
#define B_MAIL_ATTR_FLAGS	"MAIL:flags"		/* indexed int32*/
#define B_MAIL_ATTR_RECIPIENTS	"MAIL:recipients"	/* string*/
#define B_MAIL_ATTR_MIME	"MAIL:mime"		/* string*/
#define B_MAIL_ATTR_HEADER	"MAIL:header_length"	/* int32*/
#define B_MAIL_ATTR_CONTENT	"MAIL:content_length"	/* int32*/

#define B_MAIL_TO		"To: "
#define B_MAIL_CC		"Cc: "
#define B_MAIL_BCC		"Bcc: "
#define B_MAIL_FROM		"From: "
#define B_MAIL_DATE		"Date: "
#define B_MAIL_REPLY		"Reply-To: "
#define B_MAIL_SUBJECT		"Subject: "
#define B_MAIL_PRIORITY		"Priority: "
#define B_MAIL_MIME		"Mime-Version: "

/**
*
*   Parse the file and set attributes
*
**/
void SetParsedAttrs(FILE *fp)
{
   char s[16384], *ptr, *ptr1, *ptr2, *ptr3, *ptr4;
   int32 len, l, i, j, blank;
   int32 header_size, header_lines, body_size;
   time_t t;

   static struct {
      char *parse;
      char *name;
      long  id;
   } itemarray[] = {
      {B_MAIL_PRIORITY, B_MAIL_ATTR_PRIORITY, 0 },
      {B_MAIL_TO,       B_MAIL_ATTR_TO,       1 },
      {B_MAIL_CC,       B_MAIL_ATTR_CC,       2 },
      {B_MAIL_FROM,     B_MAIL_ATTR_FROM,     3 },
      {B_MAIL_SUBJECT,  B_MAIL_ATTR_SUBJECT,  4 },
      {B_MAIL_REPLY,    B_MAIL_ATTR_REPLY,    5 },
      {B_MAIL_MIME,     B_MAIL_ATTR_MIME,     6 },
      {B_MAIL_DATE,     B_MAIL_ATTR_WHEN,     7 },
      {NULL,            NULL,                 8 },
   };

   int fd = fileno(fp);
/*
*   Apparently opening the attr dir is required
*/
   DIR *d = fs_fopen_attr_dir(fd);
/*
*   Status = New
*/
   fs_write_attr(fd, B_MAIL_ATTR_STATUS, B_STRING_TYPE, 0, "New", 4);
/*
*   Flags = 0
*/
   i = 0;
   fs_write_attr(fd, B_MAIL_ATTR_FLAGS, B_INT32_TYPE, 0, &i, 4);
/*
*   Find the end of the header.
*/
   rewind(fp);

   header_lines = 0;
   header_size  = 0;
   while (fgets(s, 16384, fp)) {
/*
*   See if this is a blank line
*/
      blank = 1;
      for (i = 0; i < strlen(s); i++) {
         if ((s[i] != ' ') && (s[i] != '\t') && (s[i] != '\r') && (s[i] != '\n')) blank = 0;
      }
      if (blank) break;

      header_lines++;
   }
/*
*   Header size
*/
   header_size = ftell(fp);
   fs_write_attr(fd, B_MAIL_ATTR_HEADER, B_INT32_TYPE, 0, &header_size, 4);
/*
*   Parse the file for the various header items.
*/
   rewind(fp);
   for (i = 0; i < header_lines; i++) {
      fgets(s, 16384, fp);

      for (j = 0; itemarray[j].name; j++) {
         l = strlen(itemarray[j].parse);
         if (strncasecmp(s, itemarray[j].parse, l) == 0) {
/*
*   Special handling of case 3, From: field.
*/
            if (itemarray[j].id == 3) {
               ptr = s + l;
               while ((*ptr == ' ') || (*ptr == '\t')) ptr++;
               ptr1 = ptr;
               ptr3 = ptr;
               while ((*ptr3 != '<') && (*ptr3 != '\0') && (*ptr3 != '\r') && (*ptr3 != '\n')) ptr3++;
               ptr2 = ptr3;
               if ((*ptr3 == '\0') || (*ptr3 == '\r') || (*ptr3 == '\n')) {
                  ptr3 = ptr1;
                  ptr4 = ptr2;
               }
               else {
                  ptr2--;
                  ptr3++;
                  ptr4 = ptr3;
                  while ((*ptr4 != '>') && (*ptr4 != '\0') && (*ptr4 != '\r') && (*ptr4 != '\n')) ptr4++;
               }
               *ptr4 = '\0';
               *ptr2 = '\0';
               fs_write_attr(fd, itemarray[j].name, B_STRING_TYPE, 0, ptr3, ptr4 - ptr3 + 1);
               fs_write_attr(fd, B_MAIL_ATTR_NAME, B_STRING_TYPE, 0, ptr1, ptr2 - ptr1 + 1);
            }
/*
*   Special handling of case 7, Date: field
*/
            else if (itemarray[j].id == 7) {
               ptr = s + l;
               l = strlen(ptr) - 2;
               ptr[l] = '\0';
               t = parsedate(ptr, -1);
               if (t == -1) t = real_time_clock();
/*
*   Note: a B_TIME_TYPE should really be a bigtime_t, but in practice for
*   e-mail it turns out to be just a time_t.
*/
               fs_write_attr(fd, B_MAIL_ATTR_WHEN, B_TIME_TYPE, 0, &t, 4);
            }
            else {
               ptr = s + l;
               l = strlen(ptr) - 2;
               ptr[l] = '\0';
               fs_write_attr(fd, itemarray[j].name, B_STRING_TYPE, 0, ptr, l + 1);
            }
         }
      }
   }
/*
*   Find body size
*/
   fseek(fp, 0, SEEK_END);
   body_size = ftell(fp) - header_size;
   fs_write_attr(fd, B_MAIL_ATTR_CONTENT, B_INT32_TYPE, 0, &body_size, 4);

   if (d) fs_close_attr_dir(d);
   return;
}

/**
*
*   Function to print formatted output to the system log, and also a version
*   of sprintf.
*
**/
void dprintf(long severity, char *fmt, ...)
{
   va_list args;
   char buf[256];

   if (severity < infolevel) {
      va_start(args, fmt);
      vsprintf(buf, fmt, args);
      va_end(args);

      printf(buf);
   }
   return;
}
